import os
from typing import TypeVar, Iterable, Tuple, Set, Union, FrozenSet, MutableSet

from ordered_set import OrderedSet

from .tasks import Task


class DAG(object):
    T = TypeVar('T')
    Edge = Tuple[T, T]
    SetLike = Union[OrderedSet[T], Set[T], FrozenSet[T], MutableSet[T]]

    class NotDAGError(RuntimeError):
        pass

    def __init__(self, nodes: Iterable[T], edges: Iterable[Edge], T=Task):
        self.T = T
        self._nodes = OrderedSet(nodes)
        self._edges = OrderedSet(edges)
        if not self._is_acyclic():
            raise self.NotDAGError(f"Graph is not DAG: {self}")

    def remove_node(self, node: T) -> "DAG":
        return DAG(
            self._nodes - {node},
            OrderedSet(filter(lambda e: e[0] != node and e[1] != node, self._edges))
        )

    def remove_nodes(self, nodes: Iterable[T]) -> "DAG":
        nodes = OrderedSet(nodes)
        return DAG(
            self._nodes - nodes,
            OrderedSet(filter(lambda e: (e[0] not in nodes) and (e[1] not in nodes), self._edges))
        )

    @property
    def nodes(self) -> OrderedSet[T]:
        return self._nodes

    @property
    def edges(self) -> OrderedSet[Edge]:
        return self._edges

    @property
    def source(self):
        return next(iter(filter(lambda v: len(self.in_edges(v)) == 0, self._nodes)))

    @property
    def sink(self):
        return next(iter(filter(lambda v: len(self.out_edges(v)) == 0, self._nodes)))

    def in_edges(self, node: T) -> OrderedSet[Edge]:
        return OrderedSet(filter(lambda edge: edge[1] == node, self._edges))

    def out_edges(self, node: T) -> OrderedSet[Edge]:
        return OrderedSet(filter(lambda edge: edge[0] == node, self._edges))

    def _is_acyclic(self) -> bool:
        left = self._nodes
        while len(left) > 0:
            visited = OrderedSet()
            no_out_edges = list(filter(lambda v: len(self.out_edges(v)) == 0, left))
            if len(no_out_edges) <= 0:
                return False
            no_cycle = self.__dfs(no_out_edges[0], visited)
            if not no_cycle:
                return False
            left = left - visited
        return True

    def dependency_subgraph(self, node: T) -> "DAG":
        visited = OrderedSet()
        self.__dfs(node, visited, forward=True)
        return self.remove_nodes(self.nodes - visited)

    def __str__(self):
        return f"nodes: {','.join(str(_) for _ in self._nodes)} edges: {','.join(str(_) for _ in self._edges)}"

    def __repr__(self):
        return str(self)

    def __dfs(self, node: T, visited: SetLike, path: SetLike = None, forward=False) -> bool:
        if path is None:
            path = OrderedSet()
        visited.add(node)
        no_circle = True
        for u, v in self.in_edges(node) if not forward else self.out_edges(node):
            next_one = v if forward else u
            if next_one in path:
                no_circle = False
                continue
            if next_one in visited:
                continue
            no_circle = no_circle and self.__dfs(next_one, visited, path=path | {node}, forward=forward)
        return no_circle

    def __eq__(self, other):
        if not isinstance(other, DAG):
            return False
        return set(self.nodes) == set(other.nodes) and set(self.edges) == set(other.edges)

    def __hash__(self):
        return hash(frozenset(self.nodes) | frozenset(self.edges))

    def format_subtree(self, node: T = None) -> str:

        return os.linesep.join(
            [f'{node}'] +
            self.format_add_bar([self.__indent_formatted(self.format_subtree(v)) for u, v in self.out_edges(node)])
        )

    def format_tree(self) -> str:
        ret = os.linesep.join(
            self.format_add_bar([
                self.__indent_formatted(self.format_subtree(_)) for _ in self.nodes if len(self.in_edges(_)) == 0
            ])
        )

        return ret

    @staticmethod
    def __indent_formatted(content, level=1):
        data = content.splitlines()
        ret = list(map(
            lambda item:
            f""
            f"{(' ' if item[0] != 0 else '─') * level * 2}"
            f" {item[1]}",
            enumerate(data)
        ))
        return os.linesep.join(ret)

    @staticmethod
    def format_add_bar(lines):
        lines = sum([_.splitlines() for _ in lines], [])
        seq_indices = [idx for idx, line in enumerate(lines) if line[0] == '─']
        if len(seq_indices) <= 0:
            return lines
        first_sep, last_sep = seq_indices[0], seq_indices[-1]
        ret = []
        for idx, line in enumerate(lines):
            if first_sep <= idx < last_sep:
                adder = '├' if line[0] == '─' else '│'
            elif idx == last_sep:
                adder = '└'
            else:
                adder = ' '
            ret.append(f"{adder}{line}")
        return ret



import os
from pathlib import Path
from typing import Iterable, Callable, Tuple, Any, Set


class Task(object):
    __name_to_instance = {}

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance.__init__(*args, **kwargs)
        if instance.name in cls.__name_to_instance:
            return cls.__name_to_instance[instance.name]
        else:
            cls.__name_to_instance[instance.name] = instance
            return instance

    def __init__(
            self, name, *,
            dependencies: "Iterable[Task]" = None,
    ):
        """
        :param dependencies:
        """
        if hasattr(self, '__initialized'):
            return
        self.__initialized = True
        self._name = name
        self._children = set(dependencies) if dependencies is not None else set()  # type: Set[Task]
        self._result = None

    @staticmethod
    def task(name: str):
        return Task.__name_to_instance.get(name, None)

    @property
    def result(self) -> Any:
        return self._result

    @property
    def name(self) -> str:
        return self._name

    @property
    def dependencies(self) -> 'Set[Task]':
        return self._children

    def add_dependency(self, task: "Task"):
        self.dependencies.add(task)

    def __call__(self, *args, **kwargs) -> Tuple[bool, Any]:
        raise NotImplementedError()

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    def need_rerun(self, time: float = None) -> bool:
        if time is None:
            return True
        return any(
            _.need_rerun(time) for _ in self.dependencies
        )

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return id(self)


class ShellTask(Task):
    def __str__(self):
        return self.__str

    def __init__(
            self, command: str, name: str = None,
            dependencies: Iterable[Task] = None,
    ):
        self.__str = f"{name}: \"{command}\"" if name is not None else f"\"{command}\""
        if name is None:
            name = command
        super().__init__(
            name=name, dependencies=dependencies
        )
        self.__command = command

    def __call__(self, *args, **kwargs):
        ret = os.system(self.__command)
        self._result = ret
        return ret == 0, ret


class CallableTask(Task):
    def __str__(self):
        return self.name

    def __init__(
            self, function: Callable, name: str = None,
            dependencies: Iterable[Task] = None,
    ):
        if name is None:
            if hasattr(function, '__name__'):
                name = function.__name__
            elif hasattr(function, '__class__'):
                name = function.__class__
            else:
                name = repr(function)
        super().__init__(
            name=name, dependencies=dependencies
        )
        self.__function = function

    def __call__(self, *args, **kwargs):
        try:
            ret = self.__function(*args, **kwargs)
        except Exception as e:
            return False, e
        else:
            return True, ret


class GenerateFileTask(Task):
    def __call__(self, *args, **kwargs) -> Tuple[bool, Any]:
        return self.base()

    def __str__(self):
        return f"Generate {self.path}<-{self.base}"

    def __init__(self, path: Path, base: Task):
        """
        :param path: path to the file
        :param base: based on this task to generate the file
        """
        super().__init__(
            name=f"{str(path.resolve())}<-{base.name}",
            dependencies=base.dependencies
        )
        self.path = path
        self.base = base

    def need_rerun(self, time: float = None) -> bool:
        return (not self.path.exists()) or super().need_rerun(time)


class ReadFileTask(Task):
    def __call__(self, *args, **kwargs) -> Tuple[bool, Any]:
        return self.path.exists(), None

    def __init__(self, path: Path, dependencies: Iterable[Task] = None):
        super().__init__(
            name=f"{str(path.resolve())}",
            dependencies=dependencies
        )
        self.path = path

    def need_rerun(self, time: float = None) -> bool:
        return (not self.path.exists()) or super().need_rerun(time) or os.path.getmtime(str(self.path.resolve())) > time



from collections import defaultdict, OrderedDict
from functools import reduce
from itertools import product
from pathlib import Path
from typing import Union, Callable, Iterable, Tuple, List, Dict

from jinja2 import Template
from loguru import logger

from ..tasks import Task, ShellTask, CallableTask, ReadFileTask, GenerateFileTask

_task_builder_global_tasks = set()  # type: Set[Task]
_task_target_file_dict = {}  # type: Dict[Path, Task]
_task_dependency_file_dict = defaultdict(list)  # type: Dict[Task, List[Path]]

_drop_if_not_used = set()

reserved_keywords = [
    '__dependency__', '__d',
    '__target__', '__t',
]


def _render(
        _original: Union[None, str, Path, Task, Iterable[Union[str, Path, Task]]],
        _params
) -> Union[None, str, Path, Task, Iterable[Union[str, Path, Task]]]:
    """
    render _original
    if _original is None, return None
    if _original is a list, return a list, otherwise return a str
    :param _original:
    :param _params:
    :return:
    """
    if _original is None:
        return _original
    _is_list = True
    if isinstance(_original, (Task, str, Path)):
        _original = [_original]
        _is_list = False
    _ret = []
    for _item in _original:
        if isinstance(_item, str):
            _ret.append(Template(_item).render(
                **_params
            ))
        elif isinstance(_item, Path):
            _ret.append(Path(Template(str(_item)).render(
                **_params
            )))
        else:
            _ret.append(_item)
    if _is_list:
        return _ret
    else:
        return _ret[0]


def task(
        runner: Union[str, Callable], name: str = None,
        depend: Union[Iterable[Union[Path, str, Task]], Path, str, Task] = None,
        target: Union[Iterable[Union[str, Path]], Path, str, Task] = None,
        **template_kwargs,
) -> Task:
    depend_files, depend_tasks, is_depend_list = __parse_depend(
        _render(depend, template_kwargs)
    )
    target_files, is_target_list = __parse_targets(
        _render(target, template_kwargs)
    )
    __d = list(map(str, depend_files))
    __t = list(map(str, target_files))
    if not is_depend_list:
        __d = __d[0] if len(__d) > 0 else None
    if not is_target_list:
        __t = __t[0] if len(__t) > 0 else None

    template_kwargs.update(
        __dependency__=__d, __d=__d,
        __target__=__t, __t=__t,
    )

    if isinstance(runner, str):
        ret = ShellTask(
            command=_render(runner, template_kwargs),
            name=_render(name, template_kwargs),
            dependencies=depend_tasks
        )
    elif callable(runner):
        ret = CallableTask(
            function=runner,
            name=_render(name, template_kwargs),
            dependencies=depend_tasks
        )
    else:
        raise RuntimeError(f"runner should be a bash command or python Callable: {runner}")
    _task_dependency_file_dict[ret] = depend_files
    for _ in target_files:
        _task_target_file_dict[_] = ret
    logger.debug(f"parsed Task {ret}")
    _task_builder_global_tasks.add(ret)
    return ret


def task_product(
        runner: Union[str, Callable], name: str = None,
        depend: Iterable[Union[Path, str, Task]] = None,
        target: Iterable[Union[str, Path]] = None,
        keep=False,
        **kwargs
) -> List[Task]:
    kwargs = OrderedDict(kwargs)
    keys = list(kwargs.keys())
    values_list = list(kwargs.values())
    ret = []
    for values in product(*values_list):
        params = {k: v for k, v in zip(keys, values)}
        ret.append(task(
            runner=runner,
            name=name,
            depend=depend,
            target=target,
            **params
        ))
    if not keep:
        _drop_if_not_used.update(ret)
    return ret


def __parse_depend(
        depends: Union[Iterable[Union[str, Path, Task]], Path, str, Task] = None
) -> Tuple[List[Path], List[Task], bool]:
    depend_files, depend_tasks = [], []
    is_list = True
    if depends is None:
        return depend_files, depend_tasks, False
    if isinstance(depends, (Path, str, Task)):
        depends = [depends]
        is_list = False
    for depend in depends:
        if isinstance(depend, Path):
            depend_files.append(depend.resolve())
        elif isinstance(depend, Task):
            depend_tasks.append(depend)
        elif isinstance(depend, str):
            _task_resolved = Task.task(depend)
            if _task_resolved is not None:
                depend_tasks.append(_task_resolved)
            else:
                depend_files.append(Path(depend).resolve())
        else:
            raise RuntimeError(f"wired depend type: {depend} {type(depend)}")

    return depend_files, depend_tasks, is_list


def __parse_targets(
        targets: Union[Iterable[Union[str, Path]], str, Path] = None
) -> Tuple[List[Path], bool]:
    is_list = True
    if targets is None:
        return [], False
    if isinstance(targets, (str, Path)):
        targets = [targets]
        is_list = False
    ret = []
    for target in targets:
        if isinstance(target, Path):
            ret.append(target.resolve())
        elif isinstance(target, str):
            ret.append(Path(target).resolve())
        else:
            raise RuntimeError(f"wired target type: {target} {type(target)}")
    return ret, is_list


def _file_dependency_resolved_tasks(_tasks):
    _file_tasks = set()
    _replaced = {}
    for t in _tasks:
        dependencies_files = _task_dependency_file_dict.get(t, [])
        for f in dependencies_files:
            if f in _task_target_file_dict:
                g_task = GenerateFileTask(f, _task_target_file_dict[f])
                _file_tasks.add(g_task)
                _replaced[_task_target_file_dict[f]] = g_task
                f_task = ReadFileTask(f, [g_task])
            else:
                f_task = ReadFileTask(f)
            _file_tasks.add(f_task)
            if t in _replaced:
                _replaced[t].add_dependency(f_task)
            t.add_dependency(f_task)
    return (_tasks | _file_tasks) - set(_replaced.keys())


def _remove_unused_tasks(_tasks):
    used = reduce(
        lambda a, b: a | b,
        [_.dependencies for _ in _tasks],
        set()
    )
    return _tasks - (_drop_if_not_used - used)


def _all_built_tasks():
    tasks = _task_builder_global_tasks
    tasks = _file_dependency_resolved_tasks(tasks)
    tasks = _remove_unused_tasks(tasks)
    return tasks



from subprocess import getoutput

from .UI import *
from .tasks import Task, ShellTask, CallableTask
from .utility import str_product

__all__ = [
    "task", "task_product",
    "Task", "ShellTask", "CallableTask",
    "str_product",
    "cli"
]

__version__ = f'0.0.1.1'
__version_minor__ = f'{getoutput("git rev-parse HEAD")[:10]}@{getoutput("git rev-parse --abbrev-ref HEAD")}'



from collections import OrderedDict
from itertools import product
from typing import List

from jinja2 import Template


def str_product(template: str, **kwargs) -> List[str]:
    def apply(_template):
        return list(
            Template(str(_template)).render(
                **{k: v for k, v in zip(keys, values)}
            ) for values in product(*values_list))

    kwargs = OrderedDict(kwargs)
    keys = list(kwargs.keys())
    values_list = list(kwargs.values())
    return apply(str(template))



import codecs
import os.path
import pathlib

import pkg_resources
from setuptools import setup, find_packages


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


def parse_requirements(path):
    with pathlib.Path(path).open() as requirements_txt:
        install_requires = [
            str(requirement)
            for requirement
            in pkg_resources.parse_requirements(requirements_txt)
        ]
    return install_requires


def read_readme():
    with open('README.md', 'r') as f:
        return f.read()


setup(
    name='PyTaskScheduler',
    version=get_version('ts/__init__.py'),
    description="This package focus on build a task scheduler via Python. \
With TaskScheduler you can define a project with tasks which have some dependency on each other, and run this project.",
    author='Zeyan Li',
    author_email='li_zeyan@icloud.com',
    url='https://github.com/lizeyan/TaskScheduler',
    packages=find_packages(),
    project_urls={
        "Bug Tracker": "https://github.com/lizeyan/TaskScheduler/issues",
        "Source Code": "https://github.com/lizeyan/TaskScheduler",
    },
    install_requires=parse_requirements('requirements.txt'),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    license='MIT',
    long_description=read_readme(),
    long_description_content_type="text/markdown",
    python_requires='>=3.6',
    scripts=[
        'bin/ts'
    ]
)



import json
import os
import time
from concurrent.futures.thread import ThreadPoolExecutor
from functools import reduce
from pathlib import Path
from threading import Condition
from typing import Iterable, Union, Dict, Set

from loguru import logger

from .DAG import DAG
from .tasks import Task


class ProjectHistory(object):
    """
    Save a Project running status
    """

    def __init__(self, path=Path('.ts.history')):
        self._database = {}  # type: Dict[str, float]
        self._path = path

    @staticmethod
    def load(path: Path) -> 'ProjectHistory':
        """
        :param path: If path is not valid, a new ProjectHistory will be created
        :return:
        """
        try:
            with open(str(path.resolve()), 'r') as f:
                database = json.load(f)
        except FileNotFoundError:
            database = {}
        ret = ProjectHistory(path=path)
        ret._database = database
        return ret

    def save(self):
        """
        Save ProjectHistory to the given path
        :return:
        """
        with open(str(self._path.resolve()), 'w+') as f:
            json.dump(self._database, f, indent=4)
        return self

    def need_update(self, task: Union[Task, Path]) -> bool:
        """
        Whether a task need update based on the records status.
        :param task:
        :return:
        """
        if isinstance(task, Path):
            return not task.exists()
        if task.name not in self._database:
            return True
        task_time = self._database.get(task.name)
        return task.need_rerun(task_time)

    def set_status_(self, task: Task):
        """
        Set a task's status
        :param task:
        :return:
        """
        tic = time.time()
        self._database[task.name] = tic
        self.save()
        return self


class Project(object):
    """
    A structured container of tasks.
    """
    SUCCESS = 0
    FAIL = 1
    RUNNING = 2
    TODO = 3

    def __init__(self, tasks: Iterable[Task] = None, history_path: Path = Path('.ts.history')):
        """
        :param tasks: A iterable of tasks
        :param history_path: if path is not valid
        """
        self._history = ProjectHistory.load(history_path)
        self._tasks = set() if tasks is None else set(tasks)  # type: Set[Task]
        self._dependency_graph = None
        self._build_dependency_graph()
        self._task_finish_condition = Condition()

        self._task_status_dict = {
            task: self.TODO if self._history.need_update(task) else self.SUCCESS for task in self._tasks
        }
        self._tic_dict = {

        }

    def is_task_runnable(self, task: Task) -> bool:
        """
        A task is runnable is all its dependencies statuses are SUCCESS and its status is _TODO
        :param task:
        :return:
        """
        if any([self._task_status_dict[_] == self.FAIL for _ in task.dependencies]):
            self._task_status_dict[task] = self.FAIL
            return False
        ret = self._task_status_dict[task] == self.TODO and reduce(
            lambda a, b: a and b,
            [self._task_status_dict[_] == self.SUCCESS for _ in task.dependencies],
            True
        )
        return ret

    def find_runnable(self, todo_task_set: Set[Task]) -> Union[Task, None]:
        """
        find runnable task and remove started, failed, finished tasks
        :param todo_task_set:
        :return:
        """
        for _task in todo_task_set.copy():
            if self.is_task_runnable(_task):
                logger.debug(f"Task {_task} is runnable")
                return _task
            elif self.task_status[_task] in (self.RUNNING, self.SUCCESS, self.FAIL):
                todo_task_set.discard(_task)
        else:
            return None

    @property
    def task_status(self) -> Dict[Task, str]:
        return self._task_status_dict

    @property
    def graph(self) -> DAG:
        return self._dependency_graph

    def task_done(self, task: Task, success=True, echo_only=False):
        tic = self._tic_dict[task]
        toc = time.time()
        logger.info(f"{task} {'finished' if success else 'failed'}, elapsed time: {toc - tic:.2f}s")
        self._task_finish_condition.acquire()
        self._task_finish_condition.notify()
        self._task_finish_condition.release()
        self._task_status_dict[task] = self.SUCCESS if success else self.FAIL
        if success and not echo_only:
            self._history.set_status_(task)

    def task_started(self, task: Task):
        logger.info(f"{task} started")
        self._tic_dict[task] = time.time()
        self._task_status_dict[task] = self.RUNNING

    @property
    def tasks(self):
        return self._tasks

    def _build_dependency_graph(self):
        nodes = self._tasks
        edges = reduce(
            lambda a, b: a | b,
            [set((task, d_task) for d_task in task.dependencies) for task in self._tasks],
            set()
        )
        self._dependency_graph = DAG(nodes, edges)
        return self

    def run_task(self, task: Task, n_jobs: int, load_average: float, echo_only: bool):
        if self.task_status[task] == 'FINISHED':
            logger.debug(f"Task {task} finished already")
            return
        logger.info(f"Run Task {task} with {n_jobs} jobs, {load_average} load average")
        todo_task_set = self._dependency_graph.dependency_subgraph(task).nodes  # type: Set[Task]

        def get_job(_task: Task):
            if echo_only:
                return lambda _=_task: logger.info((str(_)))
            else:
                return lambda _=_task: _()

        if echo_only:
            def callback(_future):
                _task = future_to_task[_future]
                self.task_done(_task, echo_only=True)
        else:
            def callback(_future):
                _task = future_to_task[_future]
                _status, _ret = _future.result()
                self.task_done(_task, success=_status, echo_only=False)
                if not _status:
                    logger.error(f"Error occurs in Task {_task!s}: {_ret}")

        future_to_task = {}
        with ThreadPoolExecutor(max_workers=n_jobs if n_jobs > 0 else None) as pool:
            while len(todo_task_set) > 0:
                task = self.find_runnable(todo_task_set)
                if len(todo_task_set) <= 0:
                    break
                elif task is None:
                    self._task_finish_condition.acquire()
                    self._task_finish_condition.wait()
                    self._task_finish_condition.release()
                    continue
                while os.getloadavg()[0] > load_average:
                    time.sleep(0.5)
                self.task_started(task)
                future = pool.submit(get_job(task))
                future_to_task[future] = task
                future.add_done_callback(callback)

    def clear_history(self):
        self._history = ProjectHistory()
        self._task_status_dict = {
            task: self.TODO if self._history.need_update(task) else self.SUCCESS for task in self._tasks
        }
        return self



import os
import sys
from pathlib import Path

import click
from loguru import logger

from .task_builder import _all_built_tasks
from ..project import Project
from ..tasks import Task


@click.group(invoke_without_command=True, context_settings=dict(
    ignore_unknown_options=True,
))
@click.option(
    "--file", '-f', 'ts_file', default='task.py',
    type=click.Path(dir_okay=False),
    help='The path to TS file'
)
@click.option(
    '--working-directory', '-w', "--working-directory",
    type=click.Path(file_okay=False),
    default='.',
    help='working directory'
)
@click.option("--debug", "-d", is_flag=True, default=False)
# @click.argument('default_command_args', nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def cli(ctx, ts_file, debug, working_directory):
    os.chdir(working_directory)
    logger.remove()
    logger.add(sys.stdout, level="INFO" if not debug else "DEBUG")
    ts_file = Path(ts_file)
    with open(str(ts_file.resolve()), 'r') as f:
        exec(f.read())
    ctx.ensure_object(dict)
    ctx.obj['project'] = Project(
        tasks=_all_built_tasks(),
        history_path=ts_file.parent / ".ts.history"
    )
    ctx.obj['debug'] = debug
    if ctx.invoked_subcommand is None:
        ctx.invoke(run)
    return ctx


# noinspection PyUnusedLocal
@cli.command("run")
@click.option(
    "--jobs", '-j', 'n_jobs', default=1, type=click.IntRange(min=0),
    help='The number of concurrent jobs, and set to 0 for as many as CPUs.'
)
@click.option(
    "--load-average", '-l', 'load_average', default=float('inf'), type=click.FloatRange(min=0.0),
    help='The maximum load average allowed'
)
@click.option(
    "--echo", "-n", "echo_only", is_flag=True, default=False
)
@click.option(
    "--start-over", "-B", "start_over", is_flag=True, default=False
)
@click.argument('task', nargs=-1, type=str)
@click.pass_context
def run(ctx, n_jobs: int, load_average: float, echo_only: bool, start_over: bool, task):
    project = ctx.obj['project']  # type: Project
    debug = ctx.obj['debug']  # type: bool
    if start_over:
        logger.debug("starting over")
        project.clear_history()
    if len(task) <= 0:
        tasks = list(filter(lambda _: len(project.graph.in_edges(_)) == 0, project.graph.nodes))
    elif isinstance(task, str):
        tasks = [Task.task(task)]
    else:
        tasks = list(map(Task.task, task))
    list(map(
        lambda t: project.run_task(
            t, n_jobs=n_jobs, load_average=load_average, echo_only=echo_only
        ),
        tasks
    ))


@cli.command("tree")
@click.argument('task', nargs=-1, type=str)
@click.pass_context
def tree(ctx, task):
    project = ctx.obj['project']  # type: Project
    if len(task) <= 0:
        logger.info("tree:" + os.linesep + project.graph.format_tree())
        return
    elif isinstance(task, str):
        tasks = [Task.task(task)]
    else:
        tasks = list(map(Task.task, task))
    for task in tasks:
        logger.info(f'=' * 40)
        logger.info(f'tree Task {task}')
        logger.info(os.linesep + project.graph.format_subtree(task))


if __name__ == '__main__':
    cli()
